# NAEMO: Neighborhood-sensitive Archived Evolutionary Many-objective Optimization Algorithm
# by Raunak Sengupta, Monalisa Pal, Sriparna Saha, Sanghamitra Bandyopadhyay (2018)

# This file defines various functions required for execution of the algorithm (NAEMO)
# It imports functions from two other files as follows:
# 1. gen_func: This file contains general functions for various evolutionary algorithms
# 2. problems: This file contains the python implementation of a few benchmark problems (DTLZ1-4 problems, for example)

# This implementation is compatible with 3-, 5-, 8-, 10-, 
# 15-obj DTLZ1-4 problems

from gen_func import *
from problems import *


def mut_uni(history, nbr, v, F2, CR, min_x, max_x):
    size = len(v)
    new_pt = copy.deepcopy(v)

    a = random.sample(range(0, len(nbr)), 3)
    r1 = random.randint(0, len(history[nbr[a[0]]]) - 1)
    p1 = copy.deepcopy(history[nbr[a[0]]][r1][0])
    r2 = random.randint(0, len(history[nbr[a[1]]]) - 1)
    p2 = copy.deepcopy(history[nbr[a[1]]][r2][0])
    r3 = random.randint(0, len(history[nbr[a[2]]]) - 1)
    p3 = copy.deepcopy(history[nbr[a[2]]][r3][0])

    p = p1 + F2 * (p2 - p3)
    F = 1.5
    for i in range(0, len(new_pt)):
        mx_ = max(p[i], v[i])
        mn_ = min(p[i], v[i])
        mx = (CR * (mx_ + mn_)) + F * (0.5 * (mx_ - mn_))
        mn = (CR * (mx_ + mn_)) - F * (0.5 * (mx_ - mn_))
        r = random.random()
        new_pt[i] = min(max_x[i], max(mn + r * (mx - mn), min_x[i]))
    return new_pt


def lehmer_mean(f, p):
    return sum(f ** p) / sum(f ** (p - 1))


def generate(m1, v1, m2, v2, m3, v3):
    F1 = np.random.normal(m1, v1)
    if F1 > 1:
        F1 = 1.0
    if F1 < 0:
        F1 = 0.000001
    F2 = np.random.normal(m2, v2)
    if F2 > 1:
        F2 = 1.0
    if F2 < 0:
        F2 = 0.000001
    CR = np.random.normal(m3, v3)
    if CR > 1:
        CR = 1.0
    if CR < 0:
        CR = 0.000001
    return F1, F2, CR


def SBX_mut(history, nbr, p1, eta_c, cross_prob, min_x, max_x):
    size = len(p1)

    a = random.sample(range(1, len(nbr)), 1)
    r1 = random.randint(0, len(history[nbr[a[0]]]) - 1)
    p2 = copy.deepcopy(history[nbr[a[0]]][r1][0])

    c1, c2 = SBX(p1, p2, eta_c, cross_prob, min_x, max_x)
    v_new = []
    rnd = random.random()
    if (rnd > 0.5):
        v_new = c1
    else:
        v_new = c2
    return v_new


def diff_mut(history, nbr, v, F, CR, min_x, max_x):
    size = len(v)

    a = random.sample(range(0, len(nbr)), 3)
    r1 = random.randint(0, len(history[nbr[a[0]]]) - 1)
    p1 = copy.deepcopy(history[nbr[a[0]]][r1][0])
    r2 = random.randint(0, len(history[nbr[a[1]]]) - 1)
    p2 = copy.deepcopy(history[nbr[a[1]]][r2][0])
    r3 = random.randint(0, len(history[nbr[a[2]]]) - 1)
    p3 = copy.deepcopy(history[nbr[a[2]]][r3][0])

    k_rand = random.randint(0, size - 1)
    u = copy.deepcopy(v)
    for k in range(0, size):
        r = random.random()
        if ((r < CR) or (k == k_rand)):
            u[k] = min(max_x[k], max(p1[k] + F * (p2[k] - p3[k]), min_x[k]))
    return u


def form_ref_tab(ref_pts, nbr_size):
    table = []
    for i in range(0, len(ref_pts)):
        dist = []
        ind = []
        for j in range(0, len(ref_pts)):
            d = LA.norm(ref_pts[i] - ref_pts[j])
            dist.append(d)
            ind.append(j)
        ind = [x for (y, x) in sorted(zip(dist, ind))]
        table.append(ind[0:nbr_size])
    return table


class NAEMO(object):
    def define_exp(self, itrr, hard, soft, div, frc, fun, d_in, d_out, flg1, flg2, e_m=20):
        self.flag2 = flg2
        self.flag1 = flg1

        self.itr = itrr
        self.theta = 5.0

        self.hard_l = hard
        self.soft_l = soft
        self.divisions = div
        self.gamma = 2

        self.mut2 = "diff"
        self.mut1 = "SBX"

        self.frac = frc
        self.F_m = 0.5
        self.CR_m = 0.2
        self.eta_c_m = 30.0 / 50.0
        self.F = 0.5
        self.CR = 0.2
        self.eta_c = 30.0
        self.eta_m_m = e_m
        self.eta_m = e_m

        self.func = fun
        self.dim_in = d_in
        self.dim_out = d_out

        self.mut_prob = 1.0 / self.dim_in
        self.cross_prob = 1.0

        self.min_x = np.asarray([0] * self.dim_in)
        self.max_x = np.asarray([1] * self.dim_in)

        # factory = form_ref_pts(self.dim_out,div-1)
        # factory.form()
        # self.ref_pts = copy.deepcopy(factory.points)

        self.ref_pts = readfile(self.dim_out)
        self.IGD_pts = read_ref_pts(self.func, self.dim_out)

        self.nbr_size = 0.2
        self.k = int(self.nbr_size * len(self.ref_pts))

        self.nbrs = form_ref_tab(self.ref_pts, len(self.ref_pts) - 1)

        self.history = []
        for i in range(0, len(self.ref_pts)):
            self.history.append([])

        for i in range(0, self.gamma * self.soft_l):
            new = self.create_random_pt()
            data = associate_pt(new[1], self.ref_pts)
            d1 = LA.norm(new[1])
            g = d1 + (self.theta * data[1])
            self.history[data[0]].append([new[0], new[1], g, data[1]])

        self.intercepts = []
        self.ideal = []
        self.div_arr = []

    def create_random_pt(self):
        decision = np.zeros(self.dim_in)
        for i in range(0, self.dim_in):
            r = self.min_x[i] + (random.random() * (self.max_x[i] - self.min_x[i]))
            decision[i] = r
        objective = evaluate(self.func, decision, self.dim_out)
        return [decision, objective]

    def g(self, objective, i):
        d_2 = d2(objective, self.ref_pts[i])
        d_1 = LA.norm(objective)
        g = d_1 + (self.theta * d_2)
        return g

    def random_nbr_sample(self, arr):
        r1 = random.randint(0, len(arr) - 1)
        r2 = random.randint(0, len(self.history[arr[r1]]) - 1)
        return copy.deepcopy(self.history[arr[r1]][r2])

    def perturb(self, pt, mut, nbr_arr):
        decision = np.zeros(self.dim_in)
        if (mut == "poly_mut"):
            decision = polynomial_mutate(pt[0], self.mut_prob, self.eta_m, self.min_x, self.max_x)
        elif (mut == "SBX"):
            decision = SBX_mut(self.history, nbr_arr, pt[0], self.eta_c, self.cross_prob, self.min_x, self.max_x)
        # decision = mut_uni(self.history,nbr_arr,pt[0],0.5,0.5,self.min_x,self.max_x)
        elif (mut == "real_mut"):
            decision = real_mutate(pt[0], self.r_mut, self.min_x, self.max_x)
        else:
            decision = diff_mut(self.history, nbr_arr, pt[0], self.F, self.CR, self.min_x, self.max_x)
        # decision = mut_uni(self.history,nbr_arr,pt[0],0.5,0.5,self.min_x,self.max_x)
        objective = evaluate(self.func, decision, self.dim_out)
        return [decision, objective]

    def evaluate_opt(self):
        self.arch = []
        for i in range(0, len(self.history)):
            for j in range(0, len(self.history[i])):
                self.arch.append(self.history[i][j])
        return IGD(self.IGD_pts, self.arch)

    def dom_amt_archive(self, indv):
        l = []
        for i in range(0, len(self.history)):
            rem = []
            for j in range(0, len(self.history[i])):
                ch = check_dominate(self.history[i][j][1], indv[1])
                if ch == -1:
                    rem.append(j)
            l.append(rem)
        return l

    def remove_from_archive(self, l):
        for i in range(0, len(self.history)):
            for j in range(0, len(l[i])):
                if len(self.history[i]) > 1:
                    ind = l[i][len(l[i]) - j - 1]
                    self.history[i].pop(ind)

    def cluster(self, limit):
        pop_arr = []
        for i in range(0, len(self.history)):
            self.history[i].sort(key=lambda x: x[2])
            pop_arr.append(len(self.history[i]))
        l = sum(pop_arr)
        while l > limit:
            ind = pop_arr.index(max(pop_arr))
            self.history[ind].pop()
            pop_arr[ind] -= 1
            l -= 1

    def iterate(self, itr, gen):
        cur_nbrs = []
        cnt = 0
        j = 0

        k_val = self.k
        # k_val = ((float(self.itr - gen)/float(self.itr))*self.k)+2

        while cnt < k_val and j < len(self.nbrs[itr]):
            if len(self.history[self.nbrs[itr][j]]) > 0:
                cur_nbrs.append(self.nbrs[itr][j])
                cnt = cnt + 1
            j = j + 1
        ch_ind = itr
        while len(self.history[ch_ind]) == 0:
            a = random.randint(0, len(cur_nbrs) - 1)
            ch_ind = cur_nbrs[a]
        # ch_ind = random.randint(0,len(self.history)-1)
        r1 = random.randint(1, len(self.history[ch_ind]))
        current_pt = copy.deepcopy(self.history[ch_ind][r1 - 1])

        self.eta_c, self.F, self.CR = generate(self.eta_c_m, 0.1, self.F_m, 0.1, self.CR_m, 0.1)
        self.eta_c = 50.0 * self.eta_c

        new_pt = []
        if random.random() > self.frac:
            new_pt = self.perturb(current_pt, self.mut2, cur_nbrs)
            if (self.flag2 == True):
                new_pt = self.perturb(new_pt, "poly_mut", cur_nbrs)
        else:
            new_pt = self.perturb(current_pt, self.mut1, cur_nbrs)
            if (self.flag1 == True):
                new_pt = self.perturb(new_pt, "poly_mut", cur_nbrs)

        dom_stat = check_dominate(current_pt[1], new_pt[1]) != 1

        l = self.dom_amt_archive(new_pt)

        ret = 0
        if dom_stat == 1:
            ret = 1

            data = associate_pt(new_pt[1], self.ref_pts)
            g_val = self.g(new_pt[1], data[0])
            new_pt.append(g_val)
            new_pt.append(data[1])

            self.history[data[0]].append(new_pt)
            tot_l = 0
            tot = 0
            for i in range(0, len(l)):
                tot_l += len(l[i])
                tot += len(self.history[i])
            if (tot_l == 0):
                if tot > self.soft_l:
                    self.cluster(self.hard_l)
            else:
                self.remove_from_archive(l)
        return ret

    def optimize(self):
        self.cluster(self.hard_l)

        for i in range(0, self.itr):
            eta_s = []
            F_s = []
            CR_s = []

            for j in range(0, len(self.ref_pts)):
                a = self.iterate(j, i)
                if a == 1:
                    eta_s.append(self.eta_c)
                    F_s.append(self.F)
                    CR_s.append(self.CR)

            if len(eta_s) != 0:
                eta_s = np.array(eta_s) / 50.0
                self.eta_c_m = lehmer_mean(eta_s, 1)

            if len(F_s) != 0:
                F_s = np.array(F_s)
                self.F_m = lehmer_mean(F_s, 1)

            if len(CR_s) != 0:
                CR_s = np.array(CR_s)
                self.CR_m = lehmer_mean(CR_s, 1)

            #print(i)
            print('Progress: %0.2f%%' % (i*100/self.itr), end='\r', flush=True)
        self.cluster(self.hard_l)
